from json import JSONDecodeError
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from oauthlib.oauth2 import TokenExpiredError
from requests import Response
from requests_oauthlib import OAuth2Session

from pymfy.api.devices.category import Category
from pymfy.api.model import Command, Device, Site

BASE_URL = "https://api.somfy.com/api/v1"

SOMFY_OAUTH = "https://accounts.somfy.com/oauth/oauth/v2/auth"
SOMFY_TOKEN = "https://accounts.somfy.com/oauth/oauth/v2/token"
SOMFY_REFRESH = "https://accounts.somfy.com/oauth/oauth/v2/token"


class SomfyApi:
    def __init__(
        self,
        client_id: str,
        client_secret: str,
        redirect_uri: Optional[str] = None,
        token: Optional[Dict[str, str]] = None,
        token_updater: Optional[Callable[[str], None]] = None,
    ):

        self.client_id = client_id
        self.client_secret = client_secret
        self.token_updater = token_updater

        extra = {"client_id": self.client_id, "client_secret": self.client_secret}

        self._oauth = OAuth2Session(
            client_id=client_id,
            token=token,
            redirect_uri=redirect_uri,
            auto_refresh_kwargs=extra,
            token_updater=token_updater,
        )

    def get_sites(self) -> List[Site]:
        response = self.get("/site")
        response.raise_for_status()
        return [Site(**s) for s in response.json()]

    def get_site(self, site_id: str) -> Site:
        response = self.get("/site/" + site_id)
        response.raise_for_status()
        return Site(**response.json())

    def send_command(self, device_id: str, command: Union[Command, str]) -> str:
        if isinstance(command, str):
            command = Command(command)
        response = self.post("/device/" + device_id + "/exec", json=command)
        response.raise_for_status()
        return response.json().get("job_id")

    def get_devices(
        self, site_id: Optional[str] = None, category: Optional[Category] = None
    ) -> List[Device]:
        site_ids = [s.id for s in self.get_sites()] if site_id is None else [site_id]
        devices = []  # type: List[Device]
        for s_id in site_ids:
            response = self.get("/site/" + s_id + "/device")
            try:
                content = response.json()
            except JSONDecodeError:
                response.raise_for_status()

            if response.status_code != 200:
                # Can happen when the site does not contain any device
                continue

            devices += [
                Device(**d)
                for d in content
                if category is None or category.value in Device(**d).categories
            ]
        return devices

    def get_device(self, device_id: str) -> Device:
        response = self.get("/device/" + device_id)
        response.raise_for_status()
        return Device(**response.json())

    def get(self, path: str) -> Response:
        """Fetch a URL from the Somfy API."""
        return self._request("get", path)

    def post(self, path: str, *, json: Dict[str, Any]) -> Response:
        """Post data to the Somfy API."""
        return self._request("post", path, json=json)

    def get_authorization_url(self, state: Optional[str] = None) -> Tuple[str, str]:
        return self._oauth.authorization_url(SOMFY_OAUTH, state)

    def request_token(
        self, authorization_response: Optional[str] = None, code: Optional[str] = None
    ) -> Dict[str, str]:
        """Generic method for fetching a Somfy access token.

        :param authorization_response: Authorization response URL, the callback
                                       URL of the request back to you.
        :param code: Authorization code
        :return: A token dict
        """
        return self._oauth.fetch_token(
            SOMFY_TOKEN,
            authorization_response=authorization_response,
            code=code,
            client_secret=self.client_secret,
        )

    def refresh_tokens(self) -> Dict[str, Union[str, int]]:
        """Refresh and return new Somfy tokens."""
        token = self._oauth.refresh_token(SOMFY_REFRESH)

        if self.token_updater is not None:
            self.token_updater(token)

        return token

    def _request(self, method: str, path: str, **kwargs: Any) -> Response:
        """Make a request.

        We don't use the built-in token refresh mechanism of OAuth2 session because
        we want to allow overriding the token refresh logic.
        """
        url = BASE_URL + path
        try:
            return getattr(self._oauth, method)(url, **kwargs)
        except TokenExpiredError:
            self._oauth.token = self.refresh_tokens()

            return getattr(self._oauth, method)(url, **kwargs)



from typing import cast

from pymfy.api.devices.roller_shutter import RollerShutter
from pymfy.api.model import Command, Parameter


class Blind(RollerShutter):
    """Class to represent a blind."""

    @property
    def orientation(self) -> int:
        return cast(int, self.get_state("orientation"))

    @orientation.setter
    def orientation(self, value: int) -> None:
        command = Command("rotation", Parameter("orientation", value))
        self.send_command(command)



from typing import Union

from pymfy.api.model import Command, Device
from pymfy.api.somfy_api import SomfyApi


class SomfyDevice:
    __slots__ = "device", "api"

    def __init__(self, device: Device, api: SomfyApi):
        self.device = device
        self.api = api

    def refresh_state(self) -> None:
        self.device = self.api.get_device(self.device.id)

    def send_command(self, command: Command) -> None:
        if command.name in [capability.name for capability in self.device.capabilities]:
            self.api.send_command(self.device.id, command)
        else:
            message_template = (
                "Command {} not available. "
                "Categories: {}, Type: {}, Capabilities: {}"
            )
            message = message_template.format(
                command.name,
                self.device.categories,
                self.device.type,
                self.device.capabilities,
            )
            raise UnsupportedCommandException(message)

    def get_state(self, state_name: str) -> Union[str, int, float]:
        return next(
            (state.value for state in self.device.states if state.name == state_name)
        )


class UnsupportedCommandException(Exception):
    """Raise a Command is not listed in the capabilities of a device."""



from datetime import datetime
from enum import Enum
from typing import Optional, cast

from pymfy.api.devices.base import SomfyDevice
from pymfy.api.model import Command, Parameter


class TargetMode(Enum):
    AWAY = "away"
    AT_HOME = "at_home"
    FROST_PROTECTION = "freeze"
    MANUAL = "manuel"  # Yes in French
    SLEEP = "sleep"
    GEOFENCING = "geofencing"


class DurationType(Enum):
    FURTHER_NOTICE = "further_notice"
    NEXT_MODE = "next_mode"
    DATE = "date"


class RegulationState(Enum):
    DEROGATION = "Derogation"
    TIMETABLE = "Timetable"


class HvacState(Enum):
    HEAT = "he"
    COOL = "co"


class Thermostat(SomfyDevice):
    """Class to represent a thermostat."""

    def get_ambient_temperature(self) -> float:
        return cast(float, self.get_state("ambient_temperature"))

    def get_humidity(self) -> float:
        return cast(float, self.get_state("humidity"))

    def get_battery(self) -> int:
        return cast(int, self.get_state("battery"))

    def get_hvac_state(self) -> HvacState:
        hvac_state = self.get_state("hvac_state")
        return next((state for state in HvacState if state.value == hvac_state))

    def get_regulation_state(self) -> RegulationState:
        regulation_state = self.get_state("regulation_state")
        return next(
            (state for state in RegulationState if state.value == regulation_state)
        )

    def get_target_mode(self) -> TargetMode:
        target_mode = self.get_state("target_mode")
        return next((state for state in TargetMode if state.value == target_mode))

    def get_target_temperature(self) -> int:
        return cast(int, self.get_state("target_temperature"))

    def get_target_end_date(self) -> Optional[datetime]:
        timestamp = self.get_state("target_end_date")
        if timestamp == -1:
            return None
        return datetime.utcfromtimestamp(cast(int, timestamp))

    def get_target_start_date(self) -> Optional[datetime]:
        return datetime.utcfromtimestamp(cast(int, self.get_state("target_start_date")))

    def get_at_home_temperature(self) -> int:
        return cast(int, self.get_state("at_home_temperature"))

    def get_away_temperature(self) -> int:
        return cast(int, self.get_state("away_temperature"))

    def get_night_temperature(self) -> int:
        return cast(int, self.get_state("night_temperature"))

    def get_frost_protection_temperature(self) -> int:
        return cast(int, self.get_state("frost_protection_temperature"))

    def set_target(
        self,
        target_mode: TargetMode,
        target_temperature: int,
        duration_type: DurationType,
        duration: Optional[int] = -1,
    ) -> None:
        parameters = [
            Parameter("target_mode", target_mode.value),
            Parameter("target_temperature", target_temperature),
            Parameter("duration_type", duration_type.value),
        ]
        if duration:
            parameters.append(Parameter("duration", duration))
        command = Command("set_target", parameters)
        self.send_command(command)

    def cancel_target(self) -> None:
        self.send_command(Command("cancel_target"))

    def set_at_home_temperature(self, temperature: int) -> None:
        if abs(temperature) > 999:
            raise ValueError("temperature must be between -999 and 999")
        parameter = Parameter("at_home_temperature", temperature)
        command = Command("set_at_home_temperature", parameter)
        self.send_command(command)

    def set_away_temperature(self, temperature: int) -> None:
        if abs(temperature) > 999:
            raise ValueError("temperature must be between -999 and 999")
        parameter = Parameter("away_temperature", temperature)
        command = Command("set_away_temperature", parameter)
        self.send_command(command)

    def set_night_temperature(self, temperature: int) -> None:
        if abs(temperature) > 999:
            raise ValueError("temperature must be between -999 and 999")
        parameter = Parameter("night_temperature", temperature)
        command = Command("set_night_temperature", parameter)
        self.send_command(command)

    def set_frost_protection_temperature(self, temperature: int) -> None:
        if abs(temperature) > 999:
            raise ValueError("temperature must be between -999 and 999")
        parameter = Parameter("frost_protection_temperature", temperature)
        command = Command("set_frost_protection_temperature", parameter)
        self.send_command(command)



from typing import Any, Dict, List, Optional, Union

# pylint: disable=too-many-instance-attributes


class Site:
    __slots__ = "id", "label"

    def __init__(self, id: str, label: str):
        self.id = id
        self.label = label


class Device:
    __slots__ = (
        "id",
        "name",
        "type",
        "site_id",
        "states",
        "capabilities",
        "categories",
        "parent_id",
    )

    def __init__(
        self,
        *,
        id: str,
        type: str,
        site_id: str,
        categories: List[str],
        states: List[Dict[str, Any]],
        capabilities: List[Dict[str, Any]],
        parent_id: Optional[str] = None,
        name: Optional[str] = None,
        **_: Any
    ):
        self.id = id
        self.name = name
        self.type = type
        self.site_id = site_id
        self.categories = categories
        self.states = [State(**s) for s in states]
        self.capabilities = [Capability(**c) for c in capabilities]
        self.parent_id = parent_id


class State:
    __slots__ = "name", "value", "type"

    def __init__(self, name: str, value: Union[str, int], type: str, **_: Any):
        self.name = name
        self.value = value
        self.type = type


class Capability:
    __slots__ = "name", "parameters"

    def __init__(self, name: str, parameters: List[Dict[str, str]], **_: Any):
        self.name = name
        self.parameters = [ParameterDescription(**p) for p in parameters]


class ParameterDescription:
    __slots__ = "name", "type", "condition"

    def __init__(self, name: str, type: str, condition: Optional[str] = None, **_: Any):
        self.name = name
        self.type = type
        self.condition = condition


class Parameter(dict):  # type: ignore
    __slots__ = "name", "value"

    def __init__(self, name: str, value: Union[str, int]):
        self.name = name
        self.value = value
        dict.__init__(self, name=name, value=value)


class Command(dict):  # type: ignore
    __slots__ = "name", "parameters"

    def __init__(
        self, name: str, parameters: Union[List[Parameter], Parameter, None] = None
    ):
        if parameters is None:
            parameters = []
        if not isinstance(parameters, list):
            parameters = [parameters]
        self.name = name
        self.parameters = parameters
        dict.__init__(self, name=name, parameters=parameters)



from enum import Enum, unique


@unique
class Category(Enum):
    ROLLER_SHUTTER = "roller_shutter"
    HUB = "hub"
    INTERIOR_BLIND = "interior_blind"
    EXTERIOR_BLIND = "exterior_blind"
    CAMERA = "camera"
    HVAC = "hvac"



from typing import Optional, cast

from pymfy.api.devices.base import SomfyDevice
from pymfy.api.model import Command, Parameter


class RollerShutter(SomfyDevice):
    """Class to represent a roller shutter."""

    def get_position(self) -> int:
        return cast(int, self.get_state("position"))

    def set_position(self, value: int, low_speed: Optional[int] = False) -> None:
        command_name = "position_low_speed" if low_speed else "position"
        command = Command(command_name, Parameter("position", value))
        self.send_command(command)

    def close(self) -> None:
        self.send_command(Command("close"))

    def open(self) -> None:
        self.send_command(Command("open"))

    def stop(self) -> None:
        self.send_command(Command("stop"))

    def identify(self) -> None:
        self.send_command(Command("identify"))

    def is_closed(self) -> bool:
        return self.get_position() == 100



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pymfy', 'pymfy.api', 'pymfy.api.devices']

package_data = \
{'': ['*']}

install_requires = \
['requests-oauthlib>=1.3.0,<2.0.0']

extras_require = \
{':python_version < "3.5"': ['typing>=3.7.4,<4.0.0']}

setup_kwargs = {
    'name': 'pymfy',
    'version': '0.9.3',
    'description': 'A Somfy Open API library',
    'long_description': '<p align=center>\n    <img src="https://developer.somfy.com/sites/default/files/img/SoOpen.png"/>\n</p>\n<p align=center>\n    <a href="https://pypi.org/project/pymfy/"><img src="https://img.shields.io/pypi/v/pymfy.svg"/></a>\n    <a href="https://github.com/tetienne/somfy-open-api/actions"><img src="https://github.com/tetienne/somfy-open-api/workflows/CI/badge.svg"/></a>\n    <a href="https://codeclimate.com/github/tetienne/somfy-open-api/maintainability"><img src="https://api.codeclimate.com/v1/badges/efefe25b6c0dc796bc1c/maintainability" /></a>\n    <a href="https://codeclimate.com/github/tetienne/somfy-open-api/test_coverage"><img src="https://api.codeclimate.com/v1/badges/efefe25b6c0dc796bc1c/test_coverage" /></a>\n    <a href="https://github.com/psf/black"><img src="https://img.shields.io/badge/code%20style-black-000000.svg" /></a>\n</p>\n\nThis library is an attempt to implement the entire Somfy API in Python 3.\nDocumentation for the Somfy API can be found [here](https://developer.somfy.com/somfy-open-api/apis).\n\n\n## Get developer credentials\n\n1. Vist https://developer.somfy.com\n2. Create an account\n3. Open the *My Apps* menu\n4. Add a new App (for testing, redirect url can be anything in https)\n4. Plug in your details into the test script below.\n\n## Supported devices\nSomfy currently exposes the following type of devices:\n  - [Blinds](https://developer.somfy.com/products/blinds-interior-and-exterior)\n  - [Rolling shutters](https://developer.somfy.com/products/rolling-shutters)\n  - [Cameras](https://developer.somfy.com/products/cameras)\n  - [Connected Thermostat](https://developer.somfy.com/products/connected-thermostat)\n\nIf you find on this [page](https://developer.somfy.com/products-services-informations) devices not yet handle by this\nrepository, don\'t hesitate to open an issue.\n\n## Installation\n```\npip install pymfy\n```\n\n## Example usage\n\nPrint all covers position.\n\n```python\nimport os\nimport json\nfrom urllib.parse import urlparse, parse_qs\n\nfrom pymfy.api.devices.roller_shutter import RollerShutter\nfrom pymfy.api.somfy_api import SomfyApi\nfrom pymfy.api.devices.category import Category\n\nclient_id = r"<CLIENT_ID>"  # Consumer Key\nredir_url = "<REDIR_URL>"  # Callback URL (for testing, can be anything)\nsecret = r"<secret>"  # Consumer Secret\n\n\ndef get_token():\n    try:\n        with open(cache_path, "r") as cache:\n            return json.loads(cache.read())\n    except IOError:\n        pass\n\n\ndef set_token(token) -> None:\n    with open(cache_path, "w") as cache:\n        cache.write(json.dumps(token))\n\n\ncache_path = "/optional/cache/path"\napi = SomfyApi(client_id, secret, redir_url, token=get_token(), token_updater=set_token)\nif not os.path.isfile(cache_path):\n    authorization_url, _ = api.get_authorization_url()\n    print("Please go to {} and authorize access.".format(authorization_url))\n    authorization_response = input("Enter the full callback URL")\n    code = parse_qs(urlparse(authorization_response).query)["code"][0]\n    set_token(api.request_token(code=code))\n\ndevices = api.get_devices(category=Category.ROLLER_SHUTTER)\n\ncovers = [RollerShutter(d, api) for d in devices]\n\nfor cover in covers:\n    print(\n        "Cover {} has the following position: {}".format(\n            cover.device.name, cover.get_position()\n        )\n    )\n```\n\n## Contribute\nThe current [documentation](https://developer.somfy.com/products-services-informations) does not give enough information to implement all the devices.\nIf you want to contribute to this repository adding new devices, you can create an issue with the output of this script:\n```python\nimport json\nimport re\nfrom urllib.parse import urlparse, parse_qs\nfrom pymfy.api.somfy_api import SomfyApi\n\n\nclient_id = r"<CLIENT_ID>"  # Consumer Key\nredir_url = "<REDIR_URL>"  # Callback URL (for testing, can be anything)\nsecret = r"<secret>"  # Consumer Secret\n\napi = SomfyApi(client_id, secret, redir_url)\nauthorization_url, _ = api.get_authorization_url()\nprint("Please go to {} and authorize access.".format(authorization_url))\nauthorization_response = input("Enter the full callback URL")\ncode = parse_qs(urlparse(authorization_response).query)["code"][0]\napi.request_token(code=code)\n\nsite_ids = [s.id for s in api.get_sites()]\ndevices = [api.get("/site/" + s_id + "/device").json() for s_id in site_ids]\n\n# Remove personal information\ndumps = json.dumps(devices, sort_keys=True, indent=4, separators=(",", ": "))\ndumps = re.sub(\'".*id.*": ".*",\\n\', "", dumps)\n\nprint(dumps)\n```\n',
    'author': 'ETIENNE Thibaut',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/tetienne/somfy-open-api',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.4',
}


setup(**setup_kwargs)



from typing import cast

from pymfy.api.devices.base import SomfyDevice
from pymfy.api.model import Command


class CameraProtect(SomfyDevice):
    """Class to represent a camera"""

    def close_shutter(self) -> None:
        self.send_command(Command("shutter_close"))

    def open_shutter(self) -> None:
        self.send_command(Command("shutter_open"))

    def get_shutter_position(self) -> str:
        """ Possible returned values are opened and closed """
        return cast(str, self.get_state("shutter_position"))
